use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::Node;

/// The decision policy when determining the action in final MCTS phase
///
/// This policy drives how the MCTS algorithm chooses which action is the
/// "best" from the exploration.
#[derive(Debug)]
pub enum DecisionPolicy {
    /// Decide on the action with the most visits
    ///
    /// This option relies on the statistical confidence drive by the MCTS
    /// algorithm instead of the potentially more noisy value estimates.
    ///
    /// This is the standard policy used in most MCTS implementations, and
    /// is a good selection when not hyper-maximizing for potential gain
    MostVisits,

    /// Decide on the action with the highest average value
    ///
    /// This is non-standard, but is more aggressive in attempting to gain
    /// the highest value in a decision.
    HighestValue,
}

pub fn decide_on_action<S: GameState>(
    policy: &DecisionPolicy,
    root_node: &Node<S>,
    arena: &Arena<S>,
) -> Option<S::Action> {
    match policy {
        DecisionPolicy::MostVisits => most_visits(root_node, arena),
        DecisionPolicy::HighestValue => highest_value(root_node, arena),
    }
}

fn most_visits<S: GameState>(root_node: &Node<S>, arena: &Arena<S>) -> Option<S::Action> {
    let best_child_id: &usize = root_node
        .children
        .iter()
        .max_by(|&a, &b| {
            let node_a_visits = arena.get_node(*a).visits;
            let node_b_visits = arena.get_node(*b).visits;
            node_a_visits.partial_cmp(&node_b_visits).unwrap()
        })
        .unwrap();

    arena.get_node(*best_child_id).action.clone()
}

fn highest_value<S: GameState>(root_node: &Node<S>, arena: &Arena<S>) -> Option<S::Action> {
    let player = root_node.state.get_current_player();
    let best_child_id: &usize = root_node
        .children
        .iter()
        .max_by(|&a, &b| {
            let node_a_score = arena.get_node(*a).reward_average(player);
            let node_b_score = arena.get_node(*b).reward_average(player);
            node_a_score.partial_cmp(&node_b_score).unwrap()
        })
        .unwrap();

    arena.get_node(*best_child_id).action.clone()
}
