use std::collections::HashMap;

use crate::state::GameState;

/// The type used for reward values
pub type RewardVal = f64;

/// A node in the MCTS tree
///
/// A node represents a given game state and, using the path from the root node,
/// the actions that led to the given state. A node has a number of children
/// nodes representing the game states reachable from the given state, after
/// a given action. This creates the tree that MCTS iterates through.
///
/// This class is not thread safe, as the library does not provide for parallel
/// search.
#[derive(std::fmt::Debug)]
pub struct Node<S: GameState> {
    /// The game state at the given node, after `action`
    pub state: S,

    /// The action that led to this state from its parent
    pub action: Option<S::Action>,

    /// The identifier of the parent Node
    pub parent: Option<usize>,

    /// The number of times this node has been visited
    pub visits: u64,

    /// The player's evaluation of the node
    pub player_view: HashMap<S::Player, PlayerRewardView>,

    /// The depth of the node in the tree, this is 0 for the root node
    pub depth: usize,

    /// The identifiers of children nodes, states reachable from this one
    pub children: Vec<usize>,
}

impl<S: GameState> Node<S> {
    pub fn new(state: S, action: Option<S::Action>, parent: Option<usize>, depth: usize) -> Self {
        Node {
            state,
            action,
            parent,
            visits: 0,
            player_view: HashMap::with_capacity(2),
            depth: depth,
            children: Vec::new(),
        }
    }

    pub fn is_leaf(&self) -> bool {
        self.children.is_empty()
    }

    pub fn reward_sum(&self, player: &S::Player) -> RewardVal {
        match self.player_view.get(player) {
            Some(pv) => pv.reward_sum,
            None => 0.0,
        }
    }

    pub fn reward_average(&self, player: &S::Player) -> RewardVal {
        match self.player_view.get(player) {
            Some(pv) => pv.reward_average,
            None => 0.0,
        }
    }

    pub fn rewards(&self, player: &S::Player) -> Option<&Vec<RewardVal>> {
        match self.player_view.get(player) {
            Some(pv) => Some(&pv.rewards),
            None => None,
        }
    }

    pub fn increment_visits(&mut self) {
        self.visits += 1
    }

    pub fn record_player_reward(&mut self, player: S::Player, reward: RewardVal) {
        let pv = self.player_view.entry(player).or_default();
        pv.rewards.push(reward);
        pv.reward_sum += reward;
        pv.reward_average = pv.reward_sum / pv.rewards.len() as f64;
    }
}

/// A player's specific perspective of a node's value
///
/// Each player has their own idea of the value of a node.
pub struct PlayerRewardView {
    /// The total reward from simulations through this node
    pub reward_sum: RewardVal,

    /// The average reward from simulations through this node, often called the node value
    pub reward_average: RewardVal,

    /// The rewards we have gotten so far for simulations through this node
    pub rewards: Vec<RewardVal>,
}

impl Default for PlayerRewardView {
    fn default() -> Self {
        PlayerRewardView {
            reward_sum: 0.0,
            reward_average: 0.0,
            rewards: Vec::new(),
        }
    }
}

impl std::fmt::Debug for PlayerRewardView {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{{sum={} avg={}}}", self.reward_sum, self.reward_average)?;
        Ok(())
    }
}
